@@ -6,9 +6,7 @@
 #include "qwaylandquicksurface.h"
 #include "qwaylandinputmethodcontrol.h"
 #include "qwaylandtextinput.h"
-#if QT_WAYLAND_TEXT_INPUT_V4_WIP
-#include "qwaylandtextinputv4.h"
-#endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
+#include "qwaylandtextinputv3.h"
 #include "qwaylandqttextinputmethod.h"
 #include "qwaylandquickoutput.h"
 #include <QtWaylandCompositor/qwaylandcompositor.h>
@@ -54,65 +52,64 @@ QT_BEGIN_NAMESPACE
 static const struct {
     const char * const vertexShaderSourceFile;
     const char * const fragmentShaderSourceFile;
-    GLenum textureTarget;
     int planeCount;
     bool canProvideTexture;
     QSGMaterial::Flags materialFlags;
     QSGMaterialType materialType;
 } bufferTypes[] = {
     // BufferFormatEgl_Null
-    { "", "", 0, 0, false, {}, {} },
+    { "", "", 0, false, {}, {} },
 
-    // BufferFormatEgl_RGB
+    // BufferFormatEgl_RGB (GL_TEXTURE_2D)
     {
         ":/qt-project.org/wayland/compositor/shaders/surface.vert.qsb",
         ":/qt-project.org/wayland/compositor/shaders/surface_rgbx.frag.qsb",
-        GL_TEXTURE_2D, 1, true,
+        1, true,
         QSGMaterial::Blending,
         {}
     },
 
-    // BufferFormatEgl_RGBA
+    // BufferFormatEgl_RGBA (GL_TEXTURE_2D)
     {
         ":/qt-project.org/wayland/compositor/shaders/surface.vert.qsb",
         ":/qt-project.org/wayland/compositor/shaders/surface_rgba.frag.qsb",
-        GL_TEXTURE_2D, 1, true,
+        1, true,
         QSGMaterial::Blending,
         {}
     },
 
-    // BufferFormatEgl_EXTERNAL_OES
+    // BufferFormatEgl_EXTERNAL_OES (GL_TEXTURE_EXTERNAL_OES)
     {
         ":/qt-project.org/wayland/compositor/shaders/surface.vert.qsb",
         ":/qt-project.org/wayland/compositor/shaders/surface_oes_external.frag",
-        GL_TEXTURE_EXTERNAL_OES, 1, false,
+        1, false,
         QSGMaterial::Blending,
         {}
     },
 
-    // BufferFormatEgl_Y_U_V
+    // BufferFormatEgl_Y_U_V (GL_TEXTURE_2D)
     {
         ":/qt-project.org/wayland/compositor/shaders/surface.vert.qsb",
         ":/qt-project.org/wayland/compositor/shaders/surface_y_u_v.frag.qsb",
-        GL_TEXTURE_2D, 3, false,
+        3, false,
         QSGMaterial::Blending,
         {}
     },
 
-    // BufferFormatEgl_Y_UV
+    // BufferFormatEgl_Y_UV (GL_TEXTURE_2D)
     {
         ":/qt-project.org/wayland/compositor/shaders/surface.vert.qsb",
         ":/qt-project.org/wayland/compositor/shaders/surface_y_uv.frag.qsb",
-        GL_TEXTURE_2D, 2, false,
+        2, false,
         QSGMaterial::Blending,
         {}
     },
 
-    // BufferFormatEgl_Y_XUXV
+    // BufferFormatEgl_Y_XUXV (GL_TEXTURE_2D)
     {
         ":/qt-project.org/wayland/compositor/shaders/surface.vert.qsb",
         ":/qt-project.org/wayland/compositor/shaders/surface_y_xuxv.frag.qsb",
-        GL_TEXTURE_2D, 2, false,
+        2, false,
         QSGMaterial::Blending,
         {}
     }
@@ -160,9 +157,9 @@ void QWaylandBufferMaterialShader::setupExternalOESShader(const QString &shaderF
     descData->inVars = { texCoordInput };
 
     QShaderDescription::InOutVariable fragColorOutput;
-    texCoordInput.name = "gl_FragColor";
-    texCoordInput.type = QShaderDescription::Vec4;
-    texCoordInput.location = 0;
+    fragColorOutput.name = "gl_FragColor";
+    fragColorOutput.type = QShaderDescription::Vec4;
+    fragColorOutput.location = 0;
 
     descData->outVars = { fragColorOutput };
 
@@ -426,9 +423,38 @@ private:
     QWaylandBufferRef m_ref;
 };
 
+void QWaylandQuickItemPrivate::handleDragUpdate(QWaylandSeat *seat, const QPointF &globalPosition)
+{
+#if QT_CONFIG(draganddrop)
+    Q_Q(QWaylandQuickItem);
+    QWaylandQuickOutput *currentOutput = qobject_cast<QWaylandQuickOutput *>(q->view()->output());
+    //TODO: also check if dragging onto other outputs
+    QWaylandQuickItem *targetItem = qobject_cast<QWaylandQuickItem *>(currentOutput->pickClickableItem(q->mapToScene(globalPosition)));
+    QWaylandSurface *targetSurface = targetItem ? targetItem->surface() : nullptr;
+    if (targetSurface) {
+        QPointF position = q->mapToItem(targetItem, globalPosition);
+        QPointF surfacePosition = targetItem->mapToSurface(position);
+        seat->drag()->dragMove(targetSurface, surfacePosition);
+    }
+#else
+    Q_UNUSED(seat);
+    Q_UNUSED(globalPosition);
+#endif // QT_CONFIG(draganddrop)
+}
+
+void QWaylandQuickItemPrivate::handleDragEnded(QWaylandSeat *seat)
+{
+#if QT_CONFIG(draganddrop)
+    isDragging = false;
+    seat->drag()->drop();
+#else
+    Q_UNUSED(seat);
+#endif // QT_CONFIG(draganddrop)
+}
+
 /*!
  * \qmltype WaylandQuickItem
- * \instantiates QWaylandQuickItem
+ * \nativetype QWaylandQuickItem
  * \inqmlmodule QtWayland.Compositor
  * \since 5.8
  * \brief Provides a Qt Quick item that represents a WaylandView.
@@ -612,20 +638,9 @@ void QWaylandQuickItem::mouseMoveEvent(QMouseEvent *event)
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
-#if QT_CONFIG(draganddrop)
         if (d->isDragging) {
-            QWaylandQuickOutput *currentOutput = qobject_cast<QWaylandQuickOutput *>(view()->output());
-            //TODO: also check if dragging onto other outputs
-            QWaylandQuickItem *targetItem = qobject_cast<QWaylandQuickItem *>(currentOutput->pickClickableItem(mapToScene(event->position())));
-            QWaylandSurface *targetSurface = targetItem ? targetItem->surface() : nullptr;
-            if (targetSurface) {
-                QPointF position = mapToItem(targetItem, event->position());
-                QPointF surfacePosition = targetItem->mapToSurface(position);
-                seat->drag()->dragMove(targetSurface, surfacePosition);
-            }
-        } else
-#endif // QT_CONFIG(draganddrop)
-        {
+            d->handleDragUpdate(seat, event->position());
+        } else {
             seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->position()), event->scenePosition());
             d->hoverPos = event->position();
         }
@@ -643,15 +658,10 @@ void QWaylandQuickItem::mouseReleaseEvent(QMouseEvent *event)
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
-#if QT_CONFIG(draganddrop)
-        if (d->isDragging) {
-            d->isDragging = false;
-            seat->drag()->drop();
-        } else
-#endif
-        {
+        if (d->isDragging)
+            d->handleDragEnded(seat);
+        else
             seat->sendMouseReleaseEvent(event->button());
-        }
     } else {
         emit mouseRelease();
         event->ignore();
@@ -708,10 +718,13 @@ void QWaylandQuickItem::hoverLeaveEvent(QHoverEvent *event)
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
-        seat->setMouseFocus(nullptr);
-    } else {
-        event->ignore();
+        if (d->view.data() == seat->mouseFocus()) {
+            seat->setMouseFocus(nullptr);
+            return;
+        }
     }
+
+    event->ignore();
 }
 
 #if QT_CONFIG(wheelevent)
@@ -789,6 +802,9 @@ void QWaylandQuickItem::touchEvent(QTouchEvent *event)
             return;
         }
 
+        if (event->type() == QEvent::TouchUpdate && d->isDragging)
+            d->handleDragUpdate(seat, pointPos);
+
         event->accept();
         if (seat->mouseFocus() != d->view.data()) {
             seat->sendMouseMoveEvent(d->view.data(), pointPos, mapToScene(pointPos));
@@ -798,6 +814,8 @@ void QWaylandQuickItem::touchEvent(QTouchEvent *event)
         if (event->type() == QEvent::TouchBegin) {
             d->touchingSeats.append(seat);
         } else if (event->type() == QEvent::TouchEnd || event->type() == QEvent::TouchCancel) {
+            if (d->isDragging)
+                d->handleDragEnded(seat);
             d->touchingSeats.removeOne(seat);
         }
 
@@ -1124,13 +1142,11 @@ void QWaylandQuickItem::takeFocus(QWaylandSeat *device)
             textInput->setFocus(surface());
     }
 
-#if QT_WAYLAND_TEXT_INPUT_V4_WIP
-    if (surface()->client()->textInputProtocols().testFlag(QWaylandClient::TextInputProtocol::TextInputV4)) {
-        QWaylandTextInputV4 *textInputV4 = QWaylandTextInputV4::findIn(target);
-        if (textInputV4)
-            textInputV4->setFocus(surface());
+    if (surface()->client()->textInputProtocols().testFlag(QWaylandClient::TextInputProtocol::TextInputV3)) {
+        QWaylandTextInputV3 *textInputV3 = QWaylandTextInputV3::findIn(target);
+        if (textInputV3)
+            textInputV3->setFocus(surface());
     }
-#endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
 
     if (surface()->client()->textInputProtocols().testFlag(QWaylandClient::TextInputProtocol::QtTextInputMethodV1)) {
         QWaylandQtTextInputMethod *textInputMethod = QWaylandQtTextInputMethod::findIn(target);
